@@ -15,7 +15,8 @@ import codecs
 from collections import namedtuple
 import tempfile
 import errno
-import multiprocessing as mp
+from threading import Condition
+import socket
 
 from argparse import Namespace
 import subprocess
@@ -25,23 +26,29 @@ import getmailcore.logging
 from getmailcore.utilities import *
 
 __all__ = [
+    'run_command',
+    'ConfigurableBase',
+    'ForkingBase',
+    'ConfInstance',
+    'ConfString',
     'ConfBool',
+    'ConfInt',
+    'ConfTupleOfStrings',
+    'ConfTupleOfUnicode',
+    'ConfTupleOfTupleOfStrings',
+    'ConfPassword',
     'ConfDirectory',
     'ConfFile',
-    'ConfigurableBase',
-    'ConfInstance',
-    'ConfInt',
     'ConfMaildirPath',
     'ConfMboxPath',
-    'ConfPassword',
-    'ConfString',
-    'ConfTupleOfStrings',
-    'ConfTupleOfTupleOfStrings',
-    'ConfTupleOfUnicode',
-    'ForkingBase',
-    'run_command',
 ]
 
+if sys.version_info.major > 2:
+    TemporaryFile23 = lambda: tempfile.TemporaryFile('bw+')
+else:
+    TemporaryFile23 = lambda: tempfile.TemporaryFile('w+')
+
+
 #######################################
 def run_command(command, args):
     # Simple subprocess wrapper for running a command and fetching its exit
@@ -52,14 +59,14 @@ def run_command(command, args):
         args = list(args)
 
     # Programmer sanity checks
-    assert isinstance(command, (bytes, str)), (
+    assert isinstance(command, (bytes, unicode)), (
         'command is %s (%s)' % (command, type(command))
     )
     assert isinstance(args, list), (
         'args is %s (%s)' % (args, type(args))
     )
     for arg in args:
-        assert isinstance(arg, (bytes, str)), 'arg is %s (%s)' % (arg, type(arg))
+        assert isinstance(arg, (bytes, unicode)), 'arg is %s (%s)' % (arg, type(arg))
 
     with tempfile.TemporaryFile() as stdout, tempfile.TemporaryFile() as stderr:
 
@@ -77,7 +84,10 @@ def run_command(command, args):
         rc = p.wait()
         stdout.seek(0)
         stderr.seek(0)
-        return (rc, stdout.read().decode().strip(), stderr.read().decode().strip())
+        if sys.version_info.major == 2:
+            return (rc, stdout.read().strip(), stderr.read().strip())
+        else:
+            return (rc, stdout.read().decode().strip(), stderr.read().decode().strip())
 
 #
 # Base classes
@@ -292,7 +302,10 @@ class ConfMboxPath(ConfString):
             )
         fd = os.open(val, os.O_RDWR)
         status_old = os.fstat(fd)
-        f = os.fdopen(fd, 'br+')
+        try:
+            f = os.fdopen(fd, 'br+')
+        except ValueError: # py2
+            f = os.fdopen(fd, 'r+')
         # Check if it _is_ an mbox file.  mbox files must start with "From "
         # in their first line, or are 0-length files.
         f.seek(0, 0)
@@ -390,34 +403,64 @@ class ForkingBase(object):
         log - an object of type getmailcore.logging.Logger()
 
     '''
+    def _child_handler(self, sig, stackframe):
+        def notify():
+            self.__child_exited.acquire()
+            self.__child_exited.notify_all()
+            self.__child_exited.release()
+        self.log.trace('handler called for signal %s' % sig)
+        try:
+            pid, r = os.waitpid(self.child.childpid,0)
+        except OSError as o:
+            self.log.trace('handler called, but no children (%s)' % o)
+            notify()
+            return
+        if self.__orig_handler:
+            signal.signal(signal.SIGCHLD, self.__orig_handler)
+        self.__child_pid = pid
+        self.__child_status = r
+        self.log.trace('handler reaped child %s with status %s' % (pid, r))
+        notify()
 
-    def _wait_for_child(self, child):
-        proc = child.process
-        pid = proc.pid
-        proc.join(timeout=60)
-        # If child is still alive we joined due to timeout.
-        if (proc.is_alive()):
-            proc.terminate()
-        exitcode = proc.exitcode
-
-        if (exitcode is None):
-            raise getmailOperationError('child pid %d failed to exit' % pid)
-        if (exitcode < 0):
-            # Child killed by a signal
-            try:
-                sig = signal.Signals(-exitcode).name
-            except:
-                sig = str(-exitcode)
+    def _prepare_child(self):
+        self.log.trace('')
+        self.__child_exited = Condition()
+        self.__child_pid = 0
+        self.__child_status = None
+        self.__orig_handler = None
+        self.__orig_handler = signal.signal(signal.SIGCHLD, self._child_handler)
 
+    def _wait_for_child(self, childpid):
+        self.__child_exited.acquire()
+        if self.__child_exited.wait(socket.getdefaulttimeout() or 60) == False: # py2, <py3.2: always None
+            raise getmailOperationError('waiting child pid %d timed out'
+                                        % childpid)
+        self.__child_exited.release()
+        if self.__child_pid != childpid:
+            #self.log.error('got child pid %d, not %d' % (pid, childpid))
             raise getmailOperationError(
-                'child pid %d killed by signal %s'
-                % (pid, sig))
-
+                'got child pid %d, not %d'
+                % (self.__child_pid, childpid)
+            )
+        if os.WIFSTOPPED(self.__child_status):
+            raise getmailOperationError(
+                'child pid %d stopped by signal %d'
+                % (self.__child_pid, os.WSTOPSIG(self.__child_status))
+            )
+        if os.WIFSIGNALED(self.__child_status):
+            raise getmailOperationError(
+                'child pid %d killed by signal %d'
+                % (self.__child_pid, os.WTERMSIG(self.__child_status))
+            )
+        if not os.WIFEXITED(self.__child_status):
+            raise getmailOperationError('child pid %d failed to exit'
+                                        % self.__child_pid)
+        exitcode = os.WEXITSTATUS(self.__child_status)
         return exitcode
 
     def _pipemail(self, msg, delivered_to, received, unixfrom, stdout, stderr):
         # Write out message
-        msgfile = tempfile.TemporaryFile('bw+')
+        msgfile = TemporaryFile23()
         msgfile.write(msg.flatten(delivered_to, received, include_from=unixfrom))
         msgfile.flush()
         os.fsync(msgfile.fileno())
@@ -437,22 +480,22 @@ class ForkingBase(object):
 
     def forkchild(self, childfun, with_out=True):
         self.child = child = Namespace()
-        child.stdout = tempfile.TemporaryFile('bw+')
-        child.stderr = tempfile.TemporaryFile('bw+')
-        # XXX MacOS uses spawn instead of fork, which has difficulties to
-        # pickle/unpickle objects when multiprocessing. Set 'fork' mode by default
-        ctx = mp.get_context('fork')
-        child.process = ctx.Process(target=childfun, args=(child.stdout, child.stderr))
-        child.process.start()
-        child.childpid = child.process.pid
-        self.log.trace('spawned child %d\n' % child.childpid)
-        child.exitcode = self._wait_for_child(child)
-        child.stderr.seek(0)
-        child.err = child.stderr.read().strip().decode()
-        child.stdout.seek(0)
-        if with_out:
-            child.out = child.stdout.read().strip()
-        return child
+        child.stdout = TemporaryFile23()
+        child.stderr = TemporaryFile23()
+        child.childpid = os.fork()
+        if child.childpid != 0: # here (in the parent)
+            self._prepare_child()
+            self.log.trace('spawned child %d\n' % child.childpid)
+            child.exitcode = self._wait_for_child(child.childpid)
+            child.stderr.seek(0)
+            child.err = child.stderr.read().strip().decode()
+            child.stdout.seek(0)
+            if with_out:
+                child.out = child.stdout.read().strip()
+            return child
+        else: #== 0 in the child
+            # calls child_replace_me to execl external command
+            childfun(child.stdout, child.stderr)
 
     def get_msginfo(self, msg):
         msginfo = {}
